@@ -1,4 +1,5 @@
 import sys
+import itertools
 
 from .ndindex import NDIndex, ndindex
 from .subindex_helpers import subindex_slice
@@ -181,7 +182,7 @@ class Tuple(NDIndex):
     def raw(self):
         return tuple(i.raw for i in self.args)
 
-    def reduce(self, shape=None):
+    def reduce(self, shape=None, *, negative_int=False):
         r"""
         Reduce a Tuple index on an array of shape `shape`
 
@@ -246,8 +247,8 @@ class Tuple(NDIndex):
         Integer(1)
         >>> a[..., 1]
         array(1)
-        >>> a[1]
-        1
+        >>> a[1] # doctest: +SKIPNP1
+        np.int64(1)
 
         See https://github.com/Quansight-Labs/ndindex/issues/22.
 
@@ -278,7 +279,7 @@ class Tuple(NDIndex):
                     seen_boolean_scalar = True
                 else:
                     _args.append(s)
-            return type(self)(*_args).reduce(shape)
+            return type(self)(*_args).reduce(shape, negative_int=negative_int)
 
         arrays = []
         for i in args:
@@ -340,7 +341,7 @@ class Tuple(NDIndex):
             elif isinstance(s, BooleanArray):
                 begin_offset += s.ndim - 1
             axis = ellipsis_i - i - begin_offset
-            reduced = s.reduce(shape, axis=axis)
+            reduced = s.reduce(shape, axis=axis, negative_int=negative_int)
             if (removable
                 and isinstance(reduced, Slice)
                 and reduced == Slice(0, shape[axis], 1)):
@@ -350,7 +351,7 @@ class Tuple(NDIndex):
                 preargs.insert(0, reduced)
 
         if shape is None:
-            endargs = [s.reduce() for s in args[ellipsis_i+1:]]
+            endargs = [s.reduce(negative_int=negative_int) for s in args[ellipsis_i+1:]]
         else:
             endargs = []
             end_offset = 0
@@ -363,7 +364,7 @@ class Tuple(NDIndex):
                 if not (isinstance(s, IntegerArray) and (0 in broadcast_shape or
                                                          False in args)):
                     # Array bounds are not checked when the broadcast shape is empty
-                    s = s.reduce(shape, axis=axis)
+                    s = s.reduce(shape, axis=axis, negative_int=negative_int)
                 endargs.insert(0, s)
 
         if shape is not None:
@@ -741,6 +742,54 @@ class Tuple(NDIndex):
 
         return any(i.isempty() for i in self.args)
 
+    def selected_indices(self, shape):
+        shape = asshape(shape)
+        idx = self.expand(shape)
+
+        def _zipped_array_indices(array_indices, shape, axis=0):
+            return zip(*[i.selected_indices(shape, axis=axis+j)
+                            for j, i in enumerate(array_indices)])
+
+        def _flatten(l):
+            for element in l:
+                if isinstance(element, tuple):
+                    yield from element
+                else:
+                    yield element
+
+        # We need to zip all array indices into a single iterator.
+        iterators = []
+        array_indices = []
+        axis = 0
+        for i in idx.args:
+            if i == False:
+                return
+            elif i == True:
+                pass
+            elif isinstance(i, IntegerArray):
+                array_indices.append(i)
+            else:
+                # Tuples do not support array indices separated by slices,
+                # newaxes, or ellipses. Furthermore, if there are (non-scalar
+                # boolean) array indices, any Integer and BooleanArray indices
+                # are converted to IntegerArray. So we can assume all array
+                # indices are together in a single block, and this is the end
+                # of it.
+                if array_indices:
+                    iterators.append(_zipped_array_indices(array_indices,
+                                                           shape, axis=axis))
+                    axis += len(array_indices)
+                    array_indices.clear()
+                if i != None:
+                    iterators.append(i.selected_indices(shape, axis=axis))
+                    axis += 1
+        if idx.args and isinstance(idx.args[-1], IntegerArray):
+            iterators.append(_zipped_array_indices(array_indices,
+                                                   shape, axis=axis))
+
+        for i in itertools.product(*iterators):
+            yield Tuple(*_flatten(i)).reduce()
+
 # Imports at the bottom to avoid circular import issues
 from .array import ArrayIndex
 from .ellipsis import ellipsis
